#define __SFR_OFFSET 0
#include "avr/io.h"
/*
Rundumleuchte mit 16 LEDs mit Einzeltransistoren, mit Dimmfunktion.
Hardware: Arduino Uno mit ATmega328p, 16 LEDs, 2 Taster, 1 Potenziometer
	 ╔════════════════════════╗
	 ║                        ║
	 ║             SCL/PC5    ╟─
	 ║             SDA/PC4    ╟─
	 ║                    AREF╟─
	 ║                      00╟─
	─╢-            D13/PB5    ╟─ LED16 = Onboard-LED
	─╢5P           D12/PB4    ╟─ LED15
	─╢RES         ~D11/PB3    ╟─ LED14
	─╢3P3         ~D10/PB2    ╟─ LED13
	─╢5P           ~D9/PB1    ╟─ LED12
	─╢00            D8/PB0    ╟─ LED11
	─╢00                      ║
	─╢Uin           D7/PD7    ╟─ LED10
	 ║             ~D6/PD6    ╟─ LED9
LED1	─╢A0/PC0       ~D5/PD5    ╟─ LED8
LED2	─╢A1/PC1        D4/PD4    ╟─ LED7
LED3	─╢A2/PC2       ~D3/PD3    ╟─ LED6
LED4	─╢A3/PC3        D2/PD2    ╟─ LED5
Tasten	─╢A4/PC4        D1/PD1/TxD╟─ Arduino-Debuginterface
Poti	─╢A5/PC5        D0/PD0/RxD╟─ Arduino-Debuginterface
	 ╚════════════════════════╝

Funktion der Eingabeelemente, Vorlage = chinesisches QX-506:
* T1 = 1-Bit (im Schiebetakt) einschieben — wie bisher
* T2 = 0-Bit (im Schiebetakt) einschieben — wie bisher
* T1+T2 gemeinsam = Richtungsumkehr — neu
* Potenziometer = Geschwindigkeit — wie bisher
TODO:
Das Bitmuster und die Umlaufrichtung wird im EEPROM gespeichert.
Automatisches Standby nach ca. 2 h, Hochlauf durch Drücken von T1 oder T2.
Hochlauf der Taktfrequenz (Drehzahl) und Helligkeit beim Einschalten,
Auslauf und Ausdimmen beim Ausschalten.

Die Tasten beeinflussen A1/PC1 mit dem internen Pullup (40 kΩ):
Keine Taste:	4,90 V	0xFFC0	0xC000 ≤ summierter A/D-Wert
T1 hat 47 kΩ:	2,70 V	0x8CEC	0x8000 ≤ summierter A/D-Wert < 0xC000
T2 hat 22 kΩ:	1,81 V	0x5E79	0x5000 ≤ summierter A/D-Wert < 0x8000
T1+T2:		1,40 V	0x4912		 summierter A/D-Wert < 0x5000

Zeiten: 16 MHz / 64 (Timer0-Vorteiler) / 250 (Timer0-Periode) = 1 kHz
Umlaufzeit bei Dimmtempo „1 PWM-Schritt pro Bild“:
	1 kHz / 250 (Schritte) / 16 (LEDs) = 0,25 Hz ≘ 4 s
Per Poti langsamer oder schneller
*/
#define f_adc 2
#define f_wdt 3
#define f_dir 4
#define f_twokeys 5

.section .bss
adcAcc:		.space 4	// uint16_t acc[2];
adcSum:		.space 4	// uint16_t sum[2];
adcCnt:		.space 1	// Zähler für Summationen, bis 64
#define outAD r6
#define outAB r7
#define outAC r8
#define outBD r18
#define outBB r19
#define outBC r9
#define outCD r20
#define outCB r21
#define outCC r22
#define dim_spdL r10	// Dimm-Speed, 0x100 = 4 s Umlaufzeit
#define dim_spdH r11
#define srg_curL r12	// W12 = aktueller Leuchtzustand
#define srg_curH r13	// Beim 0-1-Übergang wird hochgedimmt
#define srg_nxtL r14	// W14 = künftiger Leuchtzustand
#define srg_nxtH r15
#define dim_phaL r16	// Dimm-Phase, H-Teil = OCR0B
#define dim_phaH r17

.section .vectors
.org 0
	rjmp	reset
// Interrupts zum Aufwecken, sonst nichts
.org 6*4
	sbi	GPIOR0,f_wdt
	reti
.org 14*4	// PWM-Zyklus (1 ms)
	rjmp	isr_oc0a

.org 15*4	// Umschalten zwischen LEDs die hochdimmen auf die, die runterdimmen
	out	PORTB,outBB
	out	PORTC,outBC
	out	PORTD,outBD
	reti
	
.org 21*4
	sei
	in	r2,SREG
	mov	r3,r24
	ldi	YL,lo8(adcAcc)
	ldi	YH,hi8(adcAcc)
	lds	r24,ADMUX
	cpi	r24,0x44
	ldi	r24,0x44
	brne	1f	// umschalten PortC4<->PortC5
	ldi	r24,0x45
	adiw	YL,2
1:	sts	ADMUX,r24
	lds	r24,ADCL
	ld	r0,Y
	add	r0,r24
	st	Y,r0
	lds	r24,ADCH
	ldd	r0,Y+1
	adc	r0,r24
	std	Y+1,r0
	cpi	YL,lo8(adcAcc+2)	// ADC5 konvertiert?
	brne	9f
	ldd	r24,Y+adcCnt-(adcAcc+2)
	subi	r24,-4
	std	Y+adcCnt-(adcAcc+2),r24
	brcs	9f
// 64 Summationen erledigt: In die Ergebnisregister kopieren und Flag setzen!
	ldi	YL,lo8(adcAcc)
	ldi	YH,hi8(adcAcc)
0:	ld	r0,Y
	std	Y+adcSum-adcAcc,r0	// kopieren und nullsetzen
	st	Y+,r1
	cpi	YL,lo8(adcAcc+4)
	brne	0b
	sbi	GPIOR0,f_adc
9:	mov	r24,r3
	out	SREG,r2
	ret
.text

reset:
	clr	r1
	ldi	r24,0x80
	sts	ACSR,r24	// Analogvergleicher aus
	ldi	r24,250-1
	out	OCR0A,r24	// Zählumfang
	lsr	r24
	out	OCR0B,r24	// Start mit PWM 50:50
	ldi	r24,0b00000011
	out	TCCR0A,r24	// PWM + verkürzter Zählumfang
	ldi	r24,8+3
	out	TCCR0B,r24	// Timer0 Vorteiler 64
	ldi	r24,0b00000110
	sts	TIMSK0,r24	// Timer0 2 Interrupts
	ldi	r24,0x3F
	sts	DIDR0,r24	// Keine digitalen Eingänge an PortC
	ldi	r24,0x44
	sts	ADMUX,r24
	ldi	r24,0b11111111	// Maximaler Taktteiler 128
	sts	ADCSRA,r24	// A/D-Wandler starten, mit Interrupt
	ldi	YL,lo8(__bss_start)
	ldi	YH,hi8(__bss_start)
0:	st	Y+,r1
	cpi	YL,lo8(__bss_end)
	brne	0b
	ldi	r24,128/3
0:	dec	r24		// 128 CPU-Takte warten
	brne	0b
	ldi	r24,0x45
	sts	ADCSRA,r24	// Multiplexer auf ADC5 umschalten
	ldi	r24,0b00111111
	out	DDRB,r24
	ldi	r24,0b00001111
	out	DDRC,r24
	ldi	r24,0b11111110
	out	DDRD,r24
	ldi	r24,hi8(F_CPU/9600-1)
	sts	UBRR0H,r24
	ldi	r24,lo8(F_CPU/9600-1)
	sts	UBRR0L,r24
	ldi	r24,0b00011000
	sts	UCSR0B,r24
	ldi	r24,0b00000001
	out	SMCR,r24	// Idle aktivieren: CPU schläft bis Interrupt
	clr	srg_curL
	clr	srg_curH
	ldi	r24,3
	mov	srg_nxtL,r24	// 2 Leuchtpunkte (TODO: Muster aus EEPROM)
	mov	srg_nxtH,r1
	ldi	r24,64
	mov	dim_spdL,r24
	mov	dim_spdH,r1	// TODO: Vom Potenziometer / aus EEPROM / hochlaufend
	sbi	GPIOR0,f_dir
	ldi	r23,4
	sei
// Der Rest läuft in Interruptroutinen
0:	sleep
	rjmp	0b


isr_oc0a:
	out	TIFR0,r23	// Gleichzeitigen, niederpriorisierten OC0B-IRQ löschen (nur bei OCR0B==OCR0A)
// Schnell hintereinander ausgeben
	out	PORTB,outAB
	out	PORTC,outAC
	out	PORTD,outAD
// PortB-Ausgabe für Compare in dieser Periode bereitstellen
	movw	outBD,outCD
	mov	outBC,outCC
	sei
// Jetzt ist Zeit für Berechnungen, und der Compare-Interrupt darf unterbrechen
// Dass der A/D-Wandler hier unterbrechen kann ist keine gute Idee,
// da adcSum atomar gelesen werden muss.
	ldi	r24,0b10100111
	sts	ADCSRA,r24	// Interrupts vom A/D-Wandler maskieren

	sbis	GPIOR0,f_adc
	 rjmp	2f
	cbi	GPIOR0,f_adc	// nach 16 MHz/128/13/2/64 ≈ 75 Hz
#if 1
	lds	dim_spdL,adcSum+2
	lds	dim_spdH,adcSum+3
	ldi	r24,6
1:	lsr	dim_spdH
	ror	dim_spdL
	dec	r24
	brne	1b
#endif

2:
	ldi	r24,0b10101111
	sts	ADCSRA,r24	// Interrupts vom A/D-Wandler zulassen
	add	dim_phaL,dim_spdL
	adc	dim_phaH,dim_spdH
	brcs	2f
	cpi	dim_phaH,250
	brcs	8f
2:	subi	dim_phaH,250	// 6 addieren
	movw	srg_curL,srg_nxtL
// Tasten auswerten (TODO!)
	//cbi	GPIOR0,f_twokeys
	//clc
	//sec
//	rjmp	1f

//t12:	clc		// T1 + T2 gedrückt: Richtung umschalten
//	sbic	GPIOR0,f_twokeys
//	 rjmp	2f	// Nicht nochmal, nur bei Flanke
//	sbi	GPIOR0,f_twokeys
// Rotationsrichtung umkehren
//	sbis	GPIOR0,f_dir
//	 rjmp	4f	// auf rechts setzen und rotieren
//	cbi	GPIOR0,f_dir
//	rjmp	3f	// links rotieren
// Je nach Rotationsrichtung Bit 0 oder Bit 15 nehmen
1:	cbi	GPIOR0,f_twokeys
	clc
2:	sbic	GPIOR0,f_dir
	 rjmp	5f	// rechts?
// Rotation auf links setzen
3:	sbrc	srg_nxtH,7
	 sec
	rjmp	7f
// Rotation auf rechts setzen
4:	sbi	GPIOR0,f_dir
5:	sbrc	srg_nxtL,0
6:	 sec
// Je nach Rotationsrichtung rechts oder links schieben
7:	sbic	GPIOR0,f_dir
	 rjmp	1f		// rechts schieben
	rol	srg_nxtL	// links schieben
	rol	srg_nxtH
	rjmp	8f
1:	ror	srg_nxtH
	ror	srg_nxtL
8:	out	OCR0B,dim_phaH
9:
	
// Port-Ausgabewerte bereitstellen: Für Compare in nächster Periode
	mov	outCC,srg_curL
	ori	outCC,0xF0// PortC fertig; Pullups an Analogeingängen
	movw	outCD,srg_curL
	lsr	outCB
	ror	outCD
	lsr	outCB	// PortB fertig
	ror	outCD
	ori	outCD,3	// PortD fertig; Pullups an RxD/TxD (falls µC solo)
// Port-Ausgabewerte bereitstellen: Für Ende dieser Periode
	mov	outAC,srg_nxtL
	ldi	r24,0xF0
	or	outAC,r24
	movw	outAD,srg_nxtL
	lsr	outAB
	ror	outAD
	lsr	outAB
	ror	outAD
	ldi	r24,3
	or	outAD,r24
	ret
	
